@@ -31,13 +31,8 @@ gboolean         gdk_display_get_debug_updates (GdkDisplay *display);
 void             gdk_display_set_debug_updates (GdkDisplay *display,
                                                 gboolean    debug_updates);
 
-void            gdk_window_move_to_rect         (GdkWindow          *window,
-                                                 const GdkRectangle *rect,
-                                                 GdkGravity          rect_anchor,
-                                                 GdkGravity          window_anchor,
-                                                 GdkAnchorHints      anchor_hints,
-                                                 gint                rect_anchor_dx,
-                                                 gint                rect_anchor_dy);
+const gchar *   gdk_get_desktop_startup_id   (void);
+const gchar *   gdk_get_desktop_autostart_id (void);
 
 typedef struct {
   /* add all private functions here, initialize them in gdk-private.c */
@@ -65,16 +60,18 @@ typedef struct {
   void             (* gdk_display_set_debug_updates) (GdkDisplay *display,
                                                       gboolean    debug_updates);
 
-  void (* gdk_window_move_to_rect) (GdkWindow          *window,
-                                    const GdkRectangle *rect,
-                                    GdkGravity          rect_anchor,
-                                    GdkGravity          window_anchor,
-                                    GdkAnchorHints      anchor_hints,
-                                    gint                rect_anchor_dx,
-                                    gint                rect_anchor_dy);
+  const gchar * (* gdk_get_desktop_startup_id)   (void);
+  const gchar * (* gdk_get_desktop_autostart_id) (void);
+
+  gboolean (* gdk_profiler_is_running) (void);
+  void     (* gdk_profiler_start)      (int fd);
+  void     (* gdk_profiler_stop)       (void);
 } GdkPrivateVTable;
 
 GDK_AVAILABLE_IN_ALL
 GdkPrivateVTable *      gdk__private__  (void);
 
+gboolean gdk_running_in_sandbox (void);
+gboolean gdk_should_use_portal (void);
+
 #endif /* __GDK__PRIVATE_H__ */
