@@ -88,9 +88,6 @@ _gdk_quartz_display_open (const gchar *display_name)
   if (_gdk_display != NULL)
     return NULL;
 
-  /* Initialize application */
-  [NSApplication sharedApplication];
-
   _gdk_display = g_object_new (gdk_quartz_display_get_type (), NULL);
   _gdk_display->device_manager = _gdk_device_manager_new (_gdk_display);
 
@@ -101,6 +98,8 @@ _gdk_quartz_display_open (const gchar *display_name)
 
   _gdk_quartz_events_init ();
 
+  /* Initialize application */
+  [NSApplication sharedApplication];
 #if 0
   /* FIXME: Remove the #if 0 when we have these functions */
   _gdk_quartz_dnd_init ();
@@ -365,8 +364,8 @@ display_reconfiguration_callback (CGDirectDisplayID            cg_display,
       return;
 
   if (flags & (kCGDisplayMovedFlag | kCGDisplayAddFlag | kCGDisplayEnabledFlag |
-               kCGDisplaySetMainFlag | kCGDisplayDesktopShapeChangedFlag |
-               kCGDisplayMirrorFlag | kCGDisplayUnMirrorFlag))
+               kCGDisplaySetMainFlag | kCGDisplayMirrorFlag |
+               kCGDisplayUnMirrorFlag))
     {
       GdkQuartzMonitor *monitor = NULL;
       guint index;
@@ -380,16 +379,17 @@ display_reconfiguration_callback (CGDirectDisplayID            cg_display,
                                   "display", display, NULL);
           monitor->id = cg_display;
           g_ptr_array_add (display->monitors, monitor);
+          display_rect (display);
+          configure_monitor (GDK_MONITOR (monitor), display);
           gdk_display_monitor_added (GDK_DISPLAY (display),
                                      GDK_MONITOR (monitor));
         }
       else
         {
           monitor = g_ptr_array_index (display->monitors, index);
+          display_rect (display);
+          configure_monitor (GDK_MONITOR (monitor), display);
         }
-
-      display_rect (display);
-      configure_monitor (GDK_MONITOR (monitor), display);
     }
   else if (flags & (kCGDisplayRemoveFlag |  kCGDisplayDisabledFlag))
     {
