@@ -20,6 +20,7 @@
 #include "gdktypes.h"
 #include "gdkprivate.h"
 #include "gdkquartz.h"
+#include "gdkinternal-quartz.h"
 
 GdkDisplay *_gdk_display = NULL;
 GdkScreen *_gdk_screen = NULL;
@@ -32,9 +33,18 @@ gdk_quartz_osx_version (void)
 
   if (minor == GDK_OSX_UNSUPPORTED)
     {
+#if MAC_OS_X_VERSION_MIN_REQUIRED < 101000
       OSErr err = Gestalt (gestaltSystemVersionMinor, (SInt32*)&minor);
 
       g_return_val_if_fail (err == noErr, GDK_OSX_UNSUPPORTED);
+#else
+      NSOperatingSystemVersion version;
+
+      version = [[NSProcessInfo processInfo] operatingSystemVersion];
+      minor = version.minorVersion;
+      if (version.majorVersion == 11)
+        minor += 16;
+#endif
     }
 
   if (minor < GDK_OSX_MIN)
