@@ -38,6 +38,9 @@
 #include <gdk/gdkcursorprivate.h>
 #include <gdk/win32/gdkwindow-win32.h>
 #include <gdk/win32/gdkwin32display.h>
+#include <gdk/win32/gdkwin32screen.h>
+#include <gdk/win32/gdkwin32keys.h>
+#include <gdk/win32/gdkselection-win32.h>
 
 #include "gdkinternals.h"
 
@@ -95,6 +98,21 @@
 #ifndef WM_MOUSEHWHEEL
 #define WM_MOUSEHWHEEL 0x20E
 #endif
+#ifndef WM_DPICHANGED
+#define WM_DPICHANGED 0x02E0
+#endif
+
+/* According to
+ * http://blog.airesoft.co.uk/2009/11/wm_messages/
+ * this is the actual internal name MS uses for this undocumented message.
+ * According to
+ * https://bugs.winehq.org/show_bug.cgi?id=15055
+ * wParam is 0
+ * lParam is a pair of virtual desktop coordinates for the popup
+ */
+#ifndef WM_SYSMENU
+#define WM_SYSMENU 0x313
+#endif
 
 #ifndef CF_DIBV5
 #define CF_DIBV5 17
@@ -183,7 +201,8 @@ HRGN	  _gdk_win32_cairo_region_to_hrgn (const cairo_region_t *region,
 					   gint                  x_origin,
 					   gint                  y_origin);
 
-cairo_region_t *_gdk_win32_hrgn_to_region    (HRGN hrgn);
+cairo_region_t *_gdk_win32_hrgn_to_region    (HRGN  hrgn,
+                                              guint scale);
 
 void	_gdk_win32_adjust_client_rect   (GdkWindow *window,
 					 RECT      *RECT);
@@ -254,20 +273,7 @@ void    _gdk_other_api_failed        (const gchar *where,
 
 extern LRESULT CALLBACK _gdk_win32_window_procedure (HWND, UINT, WPARAM, LPARAM);
 
-extern GdkWindow        *_gdk_root;
-
 extern GdkDisplay       *_gdk_display;
-extern GdkScreen        *_gdk_screen;
-
-extern gint		 _gdk_num_monitors;
-typedef struct _GdkWin32Monitor GdkWin32Monitor;
-struct _GdkWin32Monitor
-{
-  gchar *name;
-  gint width_mm, height_mm;
-  GdkRectangle rect;
-};
-extern GdkWin32Monitor  *_gdk_monitors;
 
 /* Offsets to add to Windows coordinates (which are relative to the
  * primary monitor's origin, and thus might be negative for monitors
@@ -280,6 +286,8 @@ extern HDC		 _gdk_display_hdc;
 extern HINSTANCE	 _gdk_dll_hinstance;
 extern HINSTANCE	 _gdk_app_hmodule;
 
+extern gint		 _gdk_input_ignore_core;
+
 /* These are thread specific, but GDK/win32 works OK only when invoked
  * from a single thread anyway.
  */
@@ -288,49 +296,9 @@ extern gboolean		 _gdk_input_locale_is_ime;
 extern UINT		 _gdk_input_codepage;
 
 extern guint		 _gdk_keymap_serial;
-extern gboolean		 _gdk_keyboard_has_altgr;
-extern guint		 _scancode_rshift;
 
-/* GdkAtoms: properties, targets and types */
-extern GdkAtom		 _gdk_selection;
-extern GdkAtom		 _wm_transient_for;
-extern GdkAtom		 _targets;
-extern GdkAtom		 _delete;
-extern GdkAtom		 _save_targets;
-extern GdkAtom           _utf8_string;
-extern GdkAtom		 _text;
-extern GdkAtom		 _compound_text;
-extern GdkAtom		 _text_uri_list;
-extern GdkAtom		 _text_html;
-extern GdkAtom		 _image_png;
-extern GdkAtom		 _image_jpeg;
-extern GdkAtom		 _image_bmp;
-extern GdkAtom		 _image_gif;
-
-/* DND selections */
-extern GdkAtom           _local_dnd;
-extern GdkAtom		 _gdk_win32_dropfiles;
-extern GdkAtom		 _gdk_ole2_dnd;
-
-/* Clipboard formats */
-extern UINT		 _cf_png;
-extern UINT		 _cf_jfif;
-extern UINT		 _cf_gif;
-extern UINT		 _cf_url;
-extern UINT		 _cf_html_format;
-extern UINT		 _cf_text_html;
-
-/* OLE-based DND state */
-typedef enum {
-  GDK_WIN32_DND_NONE,
-  GDK_WIN32_DND_PENDING,
-  GDK_WIN32_DND_DROPPED,
-  GDK_WIN32_DND_FAILED,
-  GDK_WIN32_DND_DRAGGING,
-} GdkWin32DndState;
-
-extern GdkWin32DndState  _dnd_target_state;
-extern GdkWin32DndState  _dnd_source_state;
+/* The singleton selection object pointer */
+extern GdkWin32Selection *_win32_selection;
 
 void _gdk_win32_dnd_do_dragdrop (void);
 void _gdk_win32_ole2_dnd_property_change (GdkAtom       type,
@@ -338,8 +306,23 @@ void _gdk_win32_ole2_dnd_property_change (GdkAtom       type,
 					  const guchar *data,
 					  gint          nelements);
 
-void  _gdk_win32_begin_modal_call (void);
-void  _gdk_win32_end_modal_call (void);
+typedef enum {
+  GDK_WIN32_MODAL_OP_NONE = 0x0,
+  GDK_WIN32_MODAL_OP_SIZE = 0x1 << 0,
+  GDK_WIN32_MODAL_OP_MOVE = 0x1 << 1,
+  GDK_WIN32_MODAL_OP_MENU = 0x1 << 2,
+  GDK_WIN32_MODAL_OP_DND  = 0x1 << 3
+} GdkWin32ModalOpKind;
+
+#define GDK_WIN32_MODAL_OP_SIZEMOVE_MASK (GDK_WIN32_MODAL_OP_SIZE | GDK_WIN32_MODAL_OP_MOVE)
+
+/* Non-zero while a modal sizing, moving, or dnd operation is in progress */
+extern GdkWin32ModalOpKind _modal_operation_in_progress;
+
+extern HWND		_modal_move_resize_window;
+
+void  _gdk_win32_begin_modal_call (GdkWin32ModalOpKind kind);
+void  _gdk_win32_end_modal_call (GdkWin32ModalOpKind kind);
 
 
 /* Options */
@@ -348,27 +331,8 @@ extern gint		 _gdk_max_colors;
 
 #define GDK_WIN32_COLORMAP_DATA(cmap) ((GdkColormapPrivateWin32 *) GDK_COLORMAP (cmap)->windowing_data)
 
-/* TRUE while a modal sizing, moving, or dnd operation is in progress */
-extern gboolean		_modal_operation_in_progress;
-
-extern HWND		_modal_move_resize_window;
-
-/* TRUE when we are emptying the clipboard ourselves */
-extern gboolean		_ignore_destroy_clipboard;
-
-/* Mapping from registered clipboard format id (native) to
- * corresponding GdkAtom
- */
-extern GHashTable	*_format_atom_table;
-
-/* Hold the result of a delayed rendering */
-extern HGLOBAL		_delayed_rendering_data;
-
 extern GdkCursor *_gdk_win32_grab_cursor;
 
-HGLOBAL _gdk_win32_selection_convert_to_dib (HGLOBAL  hdata,
-					     GdkAtom  target);
-
 /* Convert a pixbuf to an HICON (or HCURSOR).  Supports alpha under
  * Windows XP, thresholds alpha otherwise.
  */
@@ -472,6 +436,13 @@ gint      _gdk_win32_display_text_property_to_utf8_list (GdkDisplay    *display,
 							 gchar       ***list);
 gchar     *_gdk_win32_display_utf8_to_string_target (GdkDisplay *display, const gchar *str);
 
+gboolean   _gdk_win32_keymap_has_altgr           (GdkWin32Keymap *keymap);
+guint8     _gdk_win32_keymap_get_active_group    (GdkWin32Keymap *keymap);
+guint8     _gdk_win32_keymap_get_rshift_scancode (GdkWin32Keymap *keymap);
+void       _gdk_win32_keymap_set_active_layout   (GdkWin32Keymap *keymap,
+                                                  HKL             hkl);
+
+
 GdkKeymap *_gdk_win32_display_get_keymap (GdkDisplay *display);
 
 void       _gdk_win32_display_create_window_impl   (GdkDisplay    *display,
@@ -484,7 +455,7 @@ void       _gdk_win32_display_create_window_impl   (GdkDisplay    *display,
 
 /* stray GdkWindowImplWin32 members */
 void _gdk_win32_window_register_dnd (GdkWindow *window);
-GdkDragContext *_gdk_win32_window_drag_begin (GdkWindow *window, GdkDevice *device, GList *targets);
+GdkDragContext *_gdk_win32_window_drag_begin (GdkWindow *window, GdkDevice *device, GList *targets, gint x_root, gint y_root);
 gboolean _gdk_win32_window_simulate_key (GdkWindow      *window,
 				  gint            x,
 				  gint            y,
@@ -518,20 +489,8 @@ void _gdk_win32_window_change_property (GdkWindow    *window,
 void _gdk_win32_window_delete_property (GdkWindow *window, GdkAtom    property);
 
 /* Stray GdkWin32Screen members */
-GdkVisual *_gdk_win32_screen_get_system_visual (GdkScreen *screen);
-GdkVisual *_gdk_win32_screen_get_rgba_visual (GdkScreen *screen);
 gboolean _gdk_win32_screen_get_setting (GdkScreen   *screen, const gchar *name, GValue *value);
-gint _gdk_win32_screen_visual_get_best_depth (GdkScreen *screen);
-GdkVisualType _gdk_win32_screen_visual_get_best_type (GdkScreen *screen);
-GdkVisual *_gdk_win32_screen_visual_get_best (GdkScreen *screen);
-GdkVisual *_gdk_win32_screen_visual_get_best_with_depth (GdkScreen *screen, gint depth);
-GdkVisual *_gdk_win32_screen_visual_get_best_with_type (GdkScreen *screen, GdkVisualType visual_type);
-GdkVisual *_gdk_win32_screen_visual_get_best_with_both (GdkScreen *screen, gint depth, GdkVisualType visual_type);
-void _gdk_win32_screen_query_depths  (GdkScreen *screen, gint **depths, gint  *count);
-void _gdk_win32_screen_query_visual_types (GdkScreen      *screen,
-				           GdkVisualType **visual_types,
-				           gint           *count);
-GList *_gdk_win32_screen_list_visuals (GdkScreen *screen);
+void _gdk_win32_screen_on_displaychange_event (GdkWin32Screen *screen);
 
 /* Distributed display manager implementation */
 GdkDisplay *_gdk_win32_display_open (const gchar *display_name);
@@ -543,17 +502,33 @@ gchar *_gdk_win32_display_manager_get_atom_name (GdkDisplayManager *manager,
 void _gdk_win32_append_event (GdkEvent *event);
 void _gdk_win32_emit_configure_event (GdkWindow *window);
 
+
+guint32 _gdk_win32_keymap_get_decimal_mark (GdkWin32Keymap *keymap);
+
+void     _gdk_win32_window_handle_aerosnap      (GdkWindow            *window,
+                                                 GdkWin32AeroSnapCombo combo);
+
+gboolean _gdk_win32_get_window_rect             (GdkWindow  *window,
+                                                 RECT       *rect);
+void     _gdk_win32_do_emit_configure_event     (GdkWindow  *window,
+                                                 RECT        rect);
+void      gdk_win32_window_do_move_resize_drag  (GdkWindow  *window,
+                                                 gint        x,
+                                                 gint        y);
+void      gdk_win32_window_end_move_resize_drag (GdkWindow  *window);
+gboolean _gdk_win32_window_fill_min_max_info    (GdkWindow  *window,
+                                                 MINMAXINFO *mmi);
+
+gboolean _gdk_win32_window_lacks_wm_decorations (GdkWindow *window);
+
+BOOL WINAPI GtkShowWindow (GdkWindow *window,
+                           int        cmd_show);
+
+void     _gdk_win32_screen_set_font_resolution (GdkWin32Screen *win32_screen);
+
 /* Initialization */
 void _gdk_win32_windowing_init (void);
-void _gdk_windowing_window_init (GdkScreen *screen);
-void _gdk_root_window_size_init (void);
-void _gdk_monitor_init(void);
-void _gdk_visual_init (GdkScreen *screen);
 void _gdk_dnd_init    (void);
-void _gdk_events_init (void);
-void _gdk_input_init  (GdkDisplay *display);
-void _gdk_input_wintab_init_check (GdkDeviceManager *device_manager);
-
-extern gboolean _is_win8_or_later;
+void _gdk_events_init (GdkDisplay *display);
 
 #endif /* __GDK_PRIVATE_WIN32_H__ */
