@@ -64,13 +64,14 @@ gtk_container_cell_accessible_ref_child (AtkObject *obj,
 }
 
 static void
-gtk_container_cell_accessible_update_cache (GtkCellAccessible *cell)
+gtk_container_cell_accessible_update_cache (GtkCellAccessible *cell,
+                                            gboolean           emit_signal)
 {
   GtkContainerCellAccessible *container = GTK_CONTAINER_CELL_ACCESSIBLE (cell);
   GList *l;
 
   for (l = container->priv->children; l; l = l->next)
-    _gtk_cell_accessible_update_cache (l->data);
+    _gtk_cell_accessible_update_cache (l->data, emit_signal);
 }
 
 static void
@@ -82,7 +83,7 @@ gtk_container_cell_widget_set (GtkAccessible *accessible)
   for (l = container->priv->children; l; l = l->next)
     gtk_accessible_set_widget (l->data, gtk_accessible_get_widget (accessible));
 
-  GTK_ACCESSIBLE_CLASS (gtk_container_cell_accessible_parent_class)->widget_unset (accessible);
+  GTK_ACCESSIBLE_CLASS (gtk_container_cell_accessible_parent_class)->widget_set (accessible);
 }
 
 static void
@@ -110,7 +111,7 @@ gtk_container_cell_accessible_class_init (GtkContainerCellAccessibleClass *klass
   class->get_n_children = gtk_container_cell_accessible_get_n_children;
   class->ref_child = gtk_container_cell_accessible_ref_child;
 
-  accessible_class->widget_unset = gtk_container_cell_widget_set;
+  accessible_class->widget_set = gtk_container_cell_widget_set;
   accessible_class->widget_unset = gtk_container_cell_widget_unset;
 
   cell_class->update_cache = gtk_container_cell_accessible_update_cache;
@@ -141,6 +142,8 @@ gtk_container_cell_accessible_add_child (GtkContainerCellAccessible *container,
   g_return_if_fail (GTK_IS_CONTAINER_CELL_ACCESSIBLE (container));
   g_return_if_fail (GTK_IS_CELL_ACCESSIBLE (child));
 
+  g_object_ref (child);
+
   container->priv->n_children++;
   container->priv->children = g_list_append (container->priv->children, child);
   atk_object_set_parent (ATK_OBJECT (child), ATK_OBJECT (container));
@@ -156,8 +159,18 @@ gtk_container_cell_accessible_remove_child (GtkContainerCellAccessible *containe
 
   container->priv->children = g_list_remove (container->priv->children, child);
   container->priv->n_children--;
+
+  g_object_unref (child);
 }
 
+/**
+ * gtk_container_cell_accessible_get_children:
+ * @container: the container
+ *
+ * Get a list of children.
+ *
+ * Returns: (transfer none) (element-type Gtk.CellAccessible)
+ */
 GList *
 gtk_container_cell_accessible_get_children (GtkContainerCellAccessible *container)
 {
