@@ -146,7 +146,8 @@ gtk_kinetic_scrolling_init_overshoot (GtkKineticScrolling *data,
 gboolean
 gtk_kinetic_scrolling_tick (GtkKineticScrolling *data,
                             gdouble              time_delta,
-                            gdouble             *position)
+                            gdouble             *position,
+                            gdouble             *velocity)
 {
   switch(data->phase)
     {
@@ -213,6 +214,8 @@ gtk_kinetic_scrolling_tick (GtkKineticScrolling *data,
 
   if (position)
     *position = data->position;
+  if (velocity)
+    *velocity = data->velocity;
 
   return data->phase != GTK_KINETIC_SCROLLING_PHASE_FINISHED;
 }
