@@ -19,6 +19,7 @@
  * SECTION:gtkprintjob
  * @Title: GtkPrintJob
  * @Short_description: Represents a print job
+ * @Include: gtk/gtkunixprint.h
  *
  * A #GtkPrintJob object represents a job that is sent to a
  * printer. You only need to deal directly with print jobs if
@@ -26,7 +27,7 @@
  *
  * Use gtk_print_job_get_surface() to obtain the cairo surface
  * onto which the pages must be drawn. Use gtk_print_job_send()
- * to send the finished job to the printer. If you don't use cairo
+ * to send the finished job to the printer. If you don’t use cairo
  * #GtkPrintJob also supports printing of manually generated postscript,
  * via gtk_print_job_set_source_file().
  */
@@ -95,9 +96,7 @@ static void     gtk_print_job_get_property (GObject               *object,
 					    guint                  prop_id,
 					    GValue                *value,
 					    GParamSpec            *pspec);
-static GObject* gtk_print_job_constructor  (GType                  type,
-					    guint                  n_construct_properties,
-					    GObjectConstructParam *construct_params);
+static void     gtk_print_job_constructed  (GObject               *object);
 
 enum {
   STATUS_CHANGED,
@@ -124,7 +123,7 @@ gtk_print_job_class_init (GtkPrintJobClass *class)
   object_class = (GObjectClass *) class;
 
   object_class->finalize = gtk_print_job_finalize;
-  object_class->constructor = gtk_print_job_constructor;
+  object_class->constructed = gtk_print_job_constructed;
   object_class->set_property = gtk_print_job_set_property;
   object_class->get_property = gtk_print_job_get_property;
 
@@ -190,7 +189,7 @@ gtk_print_job_class_init (GtkPrintJobClass *class)
 		  G_SIGNAL_RUN_LAST,
 		  G_STRUCT_OFFSET (GtkPrintJobClass, status_changed),
 		  NULL, NULL,
-		  g_cclosure_marshal_VOID__VOID,
+		  NULL,
 		  G_TYPE_NONE, 0);
 }
 
@@ -228,23 +227,14 @@ gtk_print_job_init (GtkPrintJob *job)
 }
 
 
-static GObject*
-gtk_print_job_constructor (GType                  type,
-			   guint                  n_construct_properties,
-			   GObjectConstructParam *construct_params)
+static void
+gtk_print_job_constructed (GObject *object)
 {
-  GtkPrintJob *job;
-  GtkPrintJobPrivate *priv;
-  GObject *object;
+  GtkPrintJob *job = GTK_PRINT_JOB (object);
+  GtkPrintJobPrivate *priv = job->priv;
 
-  object =
-    G_OBJECT_CLASS (gtk_print_job_parent_class)->constructor (type,
-							      n_construct_properties,
-							      construct_params);
+  G_OBJECT_CLASS (gtk_print_job_parent_class)->constructed (object);
 
-  job = GTK_PRINT_JOB (object);
-
-  priv = job->priv;
   g_assert (priv->printer_set &&
 	    priv->settings_set &&
 	    priv->page_setup_set);
@@ -253,8 +243,6 @@ gtk_print_job_constructor (GType                  type,
 				  job,
 				  priv->settings,
 				  priv->page_setup);
-
-  return object;
 }
 
 
@@ -303,7 +291,7 @@ gtk_print_job_finalize (GObject *object)
  *
  * Creates a new #GtkPrintJob.
  *
- * Return value: a new #GtkPrintJob
+ * Returns: a new #GtkPrintJob
  *
  * Since: 2.10
  **/
@@ -329,7 +317,7 @@ gtk_print_job_new (const gchar      *title,
  * 
  * Gets the #GtkPrintSettings of the print job.
  * 
- * Return value: (transfer none): the settings of @job
+ * Returns: (transfer none): the settings of @job
  *
  * Since: 2.10
  */
@@ -347,7 +335,7 @@ gtk_print_job_get_settings (GtkPrintJob *job)
  * 
  * Gets the #GtkPrinter of the print job.
  * 
- * Return value: (transfer none): the printer of @job
+ * Returns: (transfer none): the printer of @job
  *
  * Since: 2.10
  */
@@ -365,7 +353,7 @@ gtk_print_job_get_printer (GtkPrintJob *job)
  * 
  * Gets the job title.
  * 
- * Return value: the title of @job
+ * Returns: the title of @job
  *
  * Since: 2.10
  */
@@ -383,7 +371,7 @@ gtk_print_job_get_title (GtkPrintJob *job)
  * 
  * Gets the status of the print job.
  * 
- * Return value: the status of @job
+ * Returns: the status of @job
  *
  * Since: 2.10
  */
@@ -456,6 +444,41 @@ gtk_print_job_set_source_file (GtkPrintJob *job,
   return TRUE;
 }
 
+/**
+ * gtk_print_job_set_source_fd:
+ * @job: a #GtkPrintJob
+ * @fd: a file descriptor
+ * @error: return location for errors
+ *
+ * Make the #GtkPrintJob send an existing document to the
+ * printing system. The file can be in any format understood
+ * by the platforms printing system (typically PostScript,
+ * but on many platforms PDF may work too). See
+ * gtk_printer_accepts_pdf() and gtk_printer_accepts_ps().
+ *
+ * This is similar to gtk_print_job_set_source_file(),
+ * but takes expects an open file descriptor for the file,
+ * instead of a filename.
+ *
+ * Returns: %FALSE if an error occurred
+ *
+ * Since: 3.22
+ */
+gboolean
+gtk_print_job_set_source_fd (GtkPrintJob  *job,
+                             int           fd,
+                             GError      **error)
+{
+  g_return_val_if_fail (GTK_IS_PRINT_JOB (job), FALSE);
+  g_return_val_if_fail (fd >= 0, FALSE);
+
+  job->priv->spool_io = g_io_channel_unix_new (fd);
+  if (g_io_channel_set_encoding (job->priv->spool_io, NULL, error) != G_IO_STATUS_NORMAL)
+    return FALSE;
+
+  return TRUE;
+}
+
 /**
  * gtk_print_job_get_surface:
  * @job: a #GtkPrintJob
@@ -464,7 +487,7 @@ gtk_print_job_set_source_file (GtkPrintJob *job,
  * Gets a cairo surface onto which the pages of
  * the print job should be rendered.
  * 
- * Return value: (transfer none): the cairo surface of @job
+ * Returns: (transfer none): the cairo surface of @job
  *
  * Since: 2.10
  **/
@@ -502,9 +525,10 @@ gtk_print_job_get_surface (GtkPrintJob  *job,
 
   fchmod (fd, S_IRUSR | S_IWUSR);
   
-#ifdef G_ENABLE_DEBUG 
+#ifdef G_ENABLE_DEBUG
   /* If we are debugging printing don't delete the tmp files */
-  if (!(gtk_get_debug_flags () & GTK_DEBUG_PRINTING))
+  if (GTK_DEBUG_CHECK (PRINTING)) ;
+  else
 #endif /* G_ENABLE_DEBUG */
   g_unlink (filename);
   g_free (filename);
@@ -540,7 +564,7 @@ gtk_print_job_get_surface (GtkPrintJob  *job,
  * 
  * If track_status is %TRUE, the print job will try to continue report
  * on the status of the print job in the printer queues and printer. This
- * can allow your application to show things like "out of paper" issues,
+ * can allow your application to show things like “out of paper” issues,
  * and when the print job actually reaches the printer.
  * 
  * This function is often implemented using some form of polling, so it should
@@ -575,7 +599,7 @@ gtk_print_job_set_track_print_status (GtkPrintJob *job,
  * Returns wheter jobs will be tracked after printing.
  * For details, see gtk_print_job_set_track_print_status().
  *
- * Return value: %TRUE if print job status will be reported after printing
+ * Returns: %TRUE if print job status will be reported after printing
  *
  * Since: 2.10
  */
@@ -758,7 +782,7 @@ gtk_print_job_get_page_ranges (GtkPrintJob *job,
 /**
  * gtk_print_job_set_page_ranges:
  * @job: a #GtkPrintJob
- * @ranges: (array length=n_ranges): pointer to an array of
+ * @ranges: (array length=n_ranges) (transfer full): pointer to an array of
  *    #GtkPageRange structs
  * @n_ranges: the length of the @ranges array
  *
@@ -771,6 +795,7 @@ gtk_print_job_set_page_ranges (GtkPrintJob  *job,
                                GtkPageRange *ranges,
                                gint          n_ranges)
 {
+  g_free (job->priv->page_ranges);
   job->priv->page_ranges = ranges;
   job->priv->num_page_ranges = n_ranges;
 }
