@@ -21,6 +21,8 @@
 #include "gtkmarshalers.h"
 #include "gtkprivate.h"
 #include "gtkwindowprivate.h"
+#include "gtkcssnodeprivate.h"
+#include "gtkwidgetprivate.h"
 #include "gtkintl.h"
 
 #include <gtk/gtk.h>
@@ -111,29 +113,10 @@ _gtk_text_handle_draw (GtkTextHandle         *handle,
   _gtk_text_handle_get_size (handle, &width, &height);
 
   cairo_save (cr);
-
   cairo_translate (cr, handle_window->border.left, handle_window->border.top);
 
-  gtk_style_context_save (context);
-  gtk_style_context_add_class (context,
-                               GTK_STYLE_CLASS_CURSOR_HANDLE);
-
-  if (pos == GTK_TEXT_HANDLE_POSITION_SELECTION_END)
-    {
-      gtk_style_context_add_class (context,
-                                   GTK_STYLE_CLASS_BOTTOM);
-
-      if (priv->mode == GTK_TEXT_HANDLE_MODE_CURSOR)
-        gtk_style_context_add_class (context,
-                                     GTK_STYLE_CLASS_INSERTION_CURSOR);
-    }
-  else
-    gtk_style_context_add_class (context,
-                                 GTK_STYLE_CLASS_TOP);
-
   gtk_render_handle (context, cr, 0, 0, width, height);
 
-  gtk_style_context_restore (context);
   cairo_restore (cr);
 }
 
@@ -300,6 +283,7 @@ _gtk_text_handle_ensure_widget (GtkTextHandle         *handle,
   if (!priv->windows[pos].widget)
     {
       GtkWidget *widget, *window;
+      GtkStyleContext *context;
 
       widget = gtk_event_box_new ();
       gtk_event_box_set_visible_window (GTK_EVENT_BOX (widget), TRUE);
@@ -324,8 +308,17 @@ _gtk_text_handle_ensure_widget (GtkTextHandle         *handle,
       window = gtk_widget_get_ancestor (priv->parent, GTK_TYPE_WINDOW);
       _gtk_window_add_popover (GTK_WINDOW (window), widget, priv->parent, FALSE);
 
-      gtk_style_context_set_parent (gtk_widget_get_style_context (widget),
-                                    gtk_widget_get_style_context (priv->parent));
+      context = gtk_widget_get_style_context (widget);
+      gtk_style_context_set_parent (context, gtk_widget_get_style_context (priv->parent));
+      gtk_css_node_set_name (gtk_widget_get_css_node (widget), I_("cursor-handle"));
+      if (pos == GTK_TEXT_HANDLE_POSITION_SELECTION_END)
+        {
+          gtk_style_context_add_class (context, GTK_STYLE_CLASS_BOTTOM);
+          if (priv->mode == GTK_TEXT_HANDLE_MODE_CURSOR)
+            gtk_style_context_add_class (context, GTK_STYLE_CLASS_INSERTION_CURSOR);
+        }
+      else
+        gtk_style_context_add_class (context, GTK_STYLE_CLASS_TOP);
     }
 
   return priv->windows[pos].widget;
@@ -443,6 +436,7 @@ _gtk_text_handle_update (GtkTextHandle         *handle,
 
       gtk_widget_set_size_request (handle_window->widget, width, height);
       gtk_widget_show (handle_window->widget);
+      _gtk_window_raise_popover (GTK_WINDOW (window), handle_window->widget);
       _gtk_window_set_popover_position (GTK_WINDOW (window),
                                         handle_window->widget,
                                         GTK_POS_BOTTOM, &rect);
@@ -467,25 +461,21 @@ _gtk_text_handle_set_scrollable (GtkTextHandle *handle,
 
   priv = handle->priv;
 
-  if (priv->parent_scrollable)
+  if (priv->vadj)
     {
-      if (priv->vadj)
-        {
-          g_signal_handlers_disconnect_by_data (priv->vadj, handle);
-          g_object_unref (priv->vadj);
-          priv->vadj = NULL;
-        }
-
-      if (priv->hadj)
-        {
-          g_signal_handlers_disconnect_by_data (priv->hadj, handle);
-          g_object_unref (priv->hadj);
-          priv->hadj = NULL;
-        }
- 
-      g_object_remove_weak_pointer (G_OBJECT (priv->parent_scrollable), (gpointer *) &priv->parent_scrollable);
+      g_signal_handlers_disconnect_by_data (priv->vadj, handle);
+      g_clear_object (&priv->vadj);
     }
 
+  if (priv->hadj)
+    {
+      g_signal_handlers_disconnect_by_data (priv->hadj, handle);
+      g_clear_object (&priv->hadj);
+    }
+
+  if (priv->parent_scrollable)
+    g_object_remove_weak_pointer (G_OBJECT (priv->parent_scrollable), (gpointer *) &priv->parent_scrollable);
+
   priv->parent_scrollable = scrollable;
 
   if (scrollable)
@@ -550,6 +540,21 @@ _gtk_text_handle_update_scrollable (GtkTextHandle *handle,
                         handle);
 }
 
+static GtkWidget *
+gtk_text_handle_lookup_scrollable (GtkTextHandle *handle)
+{
+  GtkTextHandlePrivate *priv;
+  GtkWidget *scrolled_window;
+
+  priv = handle->priv;
+  scrolled_window = gtk_widget_get_ancestor (priv->parent,
+                                             GTK_TYPE_SCROLLED_WINDOW);
+  if (!scrolled_window)
+    return NULL;
+
+  return gtk_bin_get_child (GTK_BIN (scrolled_window));
+}
+
 static void
 _gtk_text_handle_parent_hierarchy_changed (GtkWidget     *widget,
                                            GtkWindow     *previous_toplevel,
@@ -580,7 +585,7 @@ _gtk_text_handle_parent_hierarchy_changed (GtkWidget     *widget,
         }
     }
 
-  scrollable = gtk_widget_get_ancestor (widget, GTK_TYPE_SCROLLABLE);
+  scrollable = gtk_text_handle_lookup_scrollable (handle);
   _gtk_text_handle_update_scrollable (handle, GTK_SCROLLABLE (scrollable));
 }
 
@@ -609,7 +614,7 @@ _gtk_text_handle_set_parent (GtkTextHandle *handle,
                           G_CALLBACK (_gtk_text_handle_parent_hierarchy_changed),
                           handle);
 
-      scrollable = gtk_widget_get_ancestor (parent, GTK_TYPE_SCROLLABLE);
+      scrollable = gtk_text_handle_lookup_scrollable (handle);
     }
 
   _gtk_text_handle_update_scrollable (handle, GTK_SCROLLABLE (scrollable));
@@ -698,7 +703,7 @@ _gtk_text_handle_class_init (GtkTextHandleClass *klass)
 		  G_OBJECT_CLASS_TYPE (object_class),
 		  G_SIGNAL_RUN_LAST, 0,
 		  NULL, NULL,
-                  g_cclosure_marshal_VOID__ENUM,
+                  NULL,
                   G_TYPE_NONE, 1,
                   GTK_TYPE_TEXT_HANDLE_POSITION);
   signals[DRAG_FINISHED] =
@@ -706,7 +711,7 @@ _gtk_text_handle_class_init (GtkTextHandleClass *klass)
 		  G_OBJECT_CLASS_TYPE (object_class),
 		  G_SIGNAL_RUN_LAST, 0,
 		  NULL, NULL,
-                  g_cclosure_marshal_VOID__ENUM,
+                  NULL,
                   G_TYPE_NONE, 1,
                   GTK_TYPE_TEXT_HANDLE_POSITION);
 
@@ -770,6 +775,14 @@ _gtk_text_handle_set_mode (GtkTextHandle     *handle,
       break;
     }
 
+  if (end->widget)
+    {
+      if (mode == GTK_TEXT_HANDLE_MODE_CURSOR)
+        gtk_style_context_add_class (gtk_widget_get_style_context (end->widget), GTK_STYLE_CLASS_INSERTION_CURSOR);
+      else
+        gtk_style_context_remove_class (gtk_widget_get_style_context (end->widget), GTK_STYLE_CLASS_INSERTION_CURSOR);
+    }
+
   _gtk_text_handle_update (handle, GTK_TEXT_HANDLE_POSITION_SELECTION_START);
   _gtk_text_handle_update (handle, GTK_TEXT_HANDLE_POSITION_SELECTION_END);
 
