@@ -36,14 +36,12 @@
 #include "gtkstylecontext.h"
 #include "gtktooltipwindowprivate.h"
 #include "gtkwindowprivate.h"
+#include "gtkwidgetprivate.h"
 #include "gtkaccessible.h"
 
 #ifdef GDK_WINDOWING_WAYLAND
 #include "wayland/gdkwayland.h"
 #endif
-#ifdef GDK_WINDOWING_MIR
-#include "mir/gdkmir.h"
-#endif
 
 
 /**
@@ -113,8 +111,6 @@ struct _GtkTooltip
 
   GtkWidget *tooltip_widget;
 
-  gdouble last_x;
-  gdouble last_y;
   GdkWindow *last_window;
 
   guint timeout_id;
@@ -150,8 +146,16 @@ static void       gtk_tooltip_set_last_window      (GtkTooltip      *tooltip,
 
 static void       gtk_tooltip_handle_event_internal (GdkEvent        *event);
 
+static inline GQuark tooltip_quark (void)
+{
+  static GQuark quark;
 
-static GQuark quark_current_tooltip;
+  if G_UNLIKELY (quark == 0)
+    quark = g_quark_from_static_string ("gdk-display-current-tooltip");
+  return quark;
+}
+
+#define quark_current_tooltip tooltip_quark()
 
 G_DEFINE_TYPE (GtkTooltip, gtk_tooltip, G_TYPE_OBJECT);
 
@@ -163,8 +167,6 @@ gtk_tooltip_class_init (GtkTooltipClass *klass)
   object_class = G_OBJECT_CLASS (klass);
 
   object_class->dispose = gtk_tooltip_dispose;
-
-  quark_current_tooltip = g_quark_from_static_string ("gdk-display-current-tooltip");
 }
 
 static void
@@ -754,6 +756,12 @@ tooltip_browse_mode_expired (gpointer data)
   tooltip->browse_mode_enabled = FALSE;
   tooltip->browse_mode_timeout_id = 0;
 
+  if (tooltip->timeout_id)
+    {
+      g_source_remove (tooltip->timeout_id);
+      tooltip->timeout_id = 0;
+    }
+
   /* destroy tooltip */
   display = gtk_widget_get_display (tooltip->window);
   g_object_set_qdata (G_OBJECT (display), quark_current_tooltip, NULL);
@@ -766,6 +774,12 @@ gtk_tooltip_display_closed (GdkDisplay *display,
 			    gboolean    was_error,
 			    GtkTooltip *tooltip)
 {
+  if (tooltip->timeout_id)
+    {
+      g_source_remove (tooltip->timeout_id);
+      tooltip->timeout_id = 0;
+    }
+
   g_object_set_qdata (G_OBJECT (display), quark_current_tooltip, NULL);
 }
 
@@ -817,17 +831,10 @@ gtk_tooltip_run_requery (GtkWidget  **widget,
 
   do
     {
-      g_object_get (*widget,
-		    "has-tooltip", &has_tooltip,
-		    NULL);
+      has_tooltip = gtk_widget_get_has_tooltip (*widget);
 
       if (has_tooltip)
-	g_signal_emit_by_name (*widget,
-			       "query-tooltip",
-			       *x, *y,
-			       tooltip->keyboard_mode_enabled,
-			       tooltip,
-			       &return_value);
+        return_value = gtk_widget_query_tooltip (*widget, *x, *y, tooltip->keyboard_mode_enabled, tooltip);
 
       if (!return_value)
         {
@@ -852,229 +859,111 @@ gtk_tooltip_run_requery (GtkWidget  **widget,
   return return_value;
 }
 
-static void
-get_bounding_box (GtkWidget    *widget,
-                  GdkRectangle *bounds)
-{
-  GtkAllocation allocation;
-  GtkBorder border = { 0, };
-  GdkWindow *window;
-  gint x, y;
-  gint w, h;
-  gint x1, y1;
-  gint x2, y2;
-  gint x3, y3;
-  gint x4, y4;
-
-  window = gtk_widget_get_parent_window (widget);
-  if (window == NULL)
-    window = gtk_widget_get_window (widget);
-
-  gtk_widget_get_allocation (widget, &allocation);
-  if (GTK_IS_WINDOW (widget))
-    _gtk_window_get_shadow_width (GTK_WINDOW (widget), &border);
-  x = allocation.x + border.left;
-  y = allocation.y + border.right;
-  w = allocation.width - border.left - border.right;
-  h = allocation.height - border.top - border.bottom;
-
-  gdk_window_get_root_coords (window, x, y, &x1, &y1);
-  gdk_window_get_root_coords (window, x + w, y, &x2, &y2);
-  gdk_window_get_root_coords (window, x, y + h, &x3, &y3);
-  gdk_window_get_root_coords (window, x + w, y + h, &x4, &y4);
-
-#define MIN4(a,b,c,d) MIN(MIN(a,b),MIN(c,d))
-#define MAX4(a,b,c,d) MAX(MAX(a,b),MAX(c,d))
-
-  bounds->x = floor (MIN4 (x1, x2, x3, x4));
-  bounds->y = floor (MIN4 (y1, y2, y3, y4));
-  bounds->width = ceil (MAX4 (x1, x2, x3, x4)) - bounds->x;
-  bounds->height = ceil (MAX4 (y1, y2, y3, y4)) - bounds->y;
-}
-
 static void
 gtk_tooltip_position (GtkTooltip *tooltip,
 		      GdkDisplay *display,
-		      GtkWidget  *new_tooltip_widget)
+		      GtkWidget  *new_tooltip_widget,
+                      GdkDevice  *device)
 {
-  gint x, y, width, height;
   GdkScreen *screen;
-  gint monitor_num;
-  GdkRectangle monitor;
-  guint cursor_size;
-  GdkRectangle bounds;
-  GtkBorder border;
-
-#define MAX_DISTANCE 32
+  GtkSettings *settings;
+  GdkRectangle anchor_rect;
+  GdkWindow *window;
+  GdkWindow *widget_window;
+  GdkWindow *effective_toplevel;
+  GtkWidget *toplevel;
+  int rect_anchor_dx = 0;
+  int cursor_size;
+  int anchor_rect_padding;
 
   gtk_widget_realize (GTK_WIDGET (tooltip->current_window));
-  gtk_widget_set_visible (GTK_WIDGET (tooltip->current_window), TRUE);
+  window = _gtk_widget_get_window (GTK_WIDGET (tooltip->current_window));
 
   tooltip->tooltip_widget = new_tooltip_widget;
 
-  screen = gtk_widget_get_screen (new_tooltip_widget);
+  toplevel = _gtk_widget_get_toplevel (new_tooltip_widget);
+  gtk_widget_translate_coordinates (new_tooltip_widget, toplevel,
+                                    0, 0,
+                                    &anchor_rect.x, &anchor_rect.y);
 
-  _gtk_window_get_shadow_width (GTK_WINDOW (tooltip->current_window), &border);
+  anchor_rect.width = gtk_widget_get_allocated_width (new_tooltip_widget);
+  anchor_rect.height = gtk_widget_get_allocated_height (new_tooltip_widget);
 
-  width = gtk_widget_get_allocated_width (GTK_WIDGET (tooltip->current_window)) - border.left - border.right;
-  height = gtk_widget_get_allocated_height (GTK_WIDGET (tooltip->current_window)) - border.top - border.bottom;
+  screen = gdk_window_get_screen (window);
+  settings = gtk_settings_get_for_screen (screen);
+  g_object_get (settings,
+                "gtk-cursor-theme-size", &cursor_size,
+                NULL);
 
-  monitor_num = gdk_screen_get_monitor_at_point (screen,
-                                                 tooltip->last_x,
-                                                 tooltip->last_y);
-  gdk_screen_get_monitor_workarea (screen, monitor_num, &monitor);
+  if (cursor_size == 0)
+    cursor_size = gdk_display_get_default_cursor_size (display);
 
-  get_bounding_box (new_tooltip_widget, &bounds);
-
-  /* Position the tooltip */
-
-  cursor_size = gdk_display_get_default_cursor_size (display);
-
-  /* Try below */
-  x = bounds.x + bounds.width / 2 - width / 2;
-  y = bounds.y + bounds.height + 4;
-
-  if (y + height <= monitor.y + monitor.height)
-    {
-      if (tooltip->keyboard_mode_enabled)
-        goto found;
-
-      if (y <= tooltip->last_y + cursor_size + MAX_DISTANCE)
-        {
-          if (tooltip->last_x + cursor_size + MAX_DISTANCE < x)
-            x = tooltip->last_x + cursor_size + MAX_DISTANCE;
-          else if (x + width < tooltip->last_x - MAX_DISTANCE)
-            x = tooltip->last_x - MAX_DISTANCE - width;
-
-          goto found;
-        }
-   }
-
-  /* Try above */
-  x = bounds.x + bounds.width / 2 - width / 2;
-  y = bounds.y - height - 4;
-
-  if (y >= monitor.y)
-    {
-      if (tooltip->keyboard_mode_enabled)
-        goto found;
-
-      if (y + height >= tooltip->last_y - MAX_DISTANCE)
-        {
-          if (tooltip->last_x + cursor_size + MAX_DISTANCE < x)
-            x = tooltip->last_x + cursor_size + MAX_DISTANCE;
-          else if (x + width < tooltip->last_x - MAX_DISTANCE)
-            x = tooltip->last_x - MAX_DISTANCE - width;
-
-          goto found;
-        }
-    }
-
-  /* Try right FIXME: flip on rtl ? */
-  x = bounds.x + bounds.width + 4;
-  y = bounds.y + bounds.height / 2 - height / 2;
-
-  if (x + width <= monitor.x + monitor.width)
-    {
-      if (tooltip->keyboard_mode_enabled)
-        goto found;
-
-      if (x <= tooltip->last_x + cursor_size + MAX_DISTANCE)
-        {
-          if (tooltip->last_y + cursor_size + MAX_DISTANCE < y)
-            y = tooltip->last_y + cursor_size + MAX_DISTANCE;
-          else if (y + height < tooltip->last_y - MAX_DISTANCE)
-            y = tooltip->last_y - MAX_DISTANCE - height;
-
-          goto found;
-        }
-    }
-
-  /* Try left FIXME: flip on rtl ? */
-  x = bounds.x - width - 4;
-  y = bounds.y + bounds.height / 2 - height / 2;
-
-  if (x >= monitor.x)
-    {
-      if (tooltip->keyboard_mode_enabled)
-        goto found;
-
-      if (x + width >= tooltip->last_x - MAX_DISTANCE)
-        {
-          if (tooltip->last_y + cursor_size + MAX_DISTANCE < y)
-            y = tooltip->last_y + cursor_size + MAX_DISTANCE;
-          else if (y + height < tooltip->last_y - MAX_DISTANCE)
-            y = tooltip->last_y - MAX_DISTANCE - height;
-
-          goto found;
-        }
-    }
-
-   /* Fallback */
-  if (tooltip->keyboard_mode_enabled)
-    {
-      x = bounds.x + bounds.width / 2 - width / 2;
-      y = bounds.y + bounds.height + 4;
-    }
+  if (device)
+    anchor_rect_padding = MAX (4, cursor_size - 32);
   else
+    anchor_rect_padding = 4;
+
+  anchor_rect.x -= anchor_rect_padding;
+  anchor_rect.y -= anchor_rect_padding;
+  anchor_rect.width += anchor_rect_padding * 2;
+  anchor_rect.height += anchor_rect_padding * 2;
+
+  if (device)
     {
-      /* At cursor */
-      x = tooltip->last_x + cursor_size * 3 / 4;
-      y = tooltip->last_y + cursor_size * 3 / 4;
-    }
+      const int max_x_distance = 32;
+      /* Max 48x48 icon + default padding */
+      const int max_anchor_rect_height = 48 + 8;
+      int pointer_x, pointer_y;
 
-found:
-  /* Show it */
-  if (x + width > monitor.x + monitor.width)
-    x -= x - (monitor.x + monitor.width) + width;
-  else if (x < monitor.x)
-    x = monitor.x;
+      /*
+       * For pointer position triggered tooltips, implement the following
+       * semantics:
+       *
+       * If the anchor rectangle is too tall (meaning if we'd be constrained
+       * and flip, it'd flip too far away), rely only on the pointer position
+       * to position the tooltip. The approximate pointer cursorrectangle is
+       * used as a anchor rectantgle.
+       *
+       * If the anchor rectangle isn't to tall, make sure the tooltip isn't too
+       * far away from the pointer position.
+       */
+      widget_window = _gtk_widget_get_window (new_tooltip_widget);
+      effective_toplevel = gdk_window_get_effective_toplevel (widget_window);
+      gdk_window_get_device_position (effective_toplevel,
+                                      device,
+                                      &pointer_x, &pointer_y, NULL);
 
-  if (y + height > monitor.y + monitor.height)
-    y -= y - (monitor.y + monitor.height) + height;
-  else if (y < monitor.y)
-    y = monitor.y;
-
-  if (!tooltip->keyboard_mode_enabled)
-    {
-      /* don't pop up under the pointer */
-      if (x <= tooltip->last_x && tooltip->last_x < x + width &&
-          y <= tooltip->last_y && tooltip->last_y < y + height)
-        y = tooltip->last_y - height - 2;
-    }
-
-#ifdef GDK_WINDOWING_WAYLAND
-  /* set the transient parent on the tooltip when running with the Wayland
-   * backend to allow correct positioning of the tooltip windows
-   */
-  if (GDK_IS_WAYLAND_DISPLAY (display))
-    {
-      GtkWidget *toplevel;
-
-      toplevel = gtk_widget_get_toplevel (tooltip->tooltip_widget);
-      if (GTK_IS_WINDOW (toplevel))
-        gtk_window_set_transient_for (GTK_WINDOW (tooltip->current_window),
-                                      GTK_WINDOW (toplevel));
-    }
-#endif
-#ifdef GDK_WINDOWING_MIR
-      /* Set the transient parent on the tooltip when running with the Mir
-       * backend to allow correct positioning of the tooltip windows */
-      if (GDK_IS_MIR_DISPLAY (display))
+      if (anchor_rect.height > max_anchor_rect_height)
         {
-          GtkWidget *toplevel;
-
-          toplevel = gtk_widget_get_toplevel (tooltip->tooltip_widget);
-          if (GTK_IS_WINDOW (toplevel))
-            gtk_window_set_transient_for (GTK_WINDOW (tooltip->current_window),
-                                          GTK_WINDOW (toplevel));
+          anchor_rect.x = pointer_x - 4;
+          anchor_rect.y = pointer_y - 4;
+          anchor_rect.width = cursor_size;
+          anchor_rect.height = cursor_size;
         }
-#endif
+      else
+        {
+          int anchor_point_x;
+          int x_distance;
 
-  x -= border.left;
-  y -= border.top;
+          anchor_point_x = anchor_rect.x + anchor_rect.width / 2;
+          x_distance = pointer_x - anchor_point_x;
 
-  gtk_window_move (GTK_WINDOW (tooltip->current_window), x, y);
+          if (x_distance > max_x_distance)
+            rect_anchor_dx = x_distance - max_x_distance;
+          else if (x_distance < -max_x_distance)
+            rect_anchor_dx = x_distance + max_x_distance;
+        }
+    }
+
+  gtk_window_set_transient_for (GTK_WINDOW (tooltip->current_window),
+                                GTK_WINDOW (toplevel));
+
+  gdk_window_move_to_rect (window,
+                           &anchor_rect,
+                           GDK_GRAVITY_SOUTH,
+                           GDK_GRAVITY_NORTH,
+                           GDK_ANCHOR_FLIP_Y | GDK_ANCHOR_SLIDE_X,
+                           rect_anchor_dx, 0);
   gtk_widget_show (GTK_WIDGET (tooltip->current_window));
 }
 
@@ -1083,7 +972,7 @@ gtk_tooltip_show_tooltip (GdkDisplay *display)
 {
   gint x, y;
   GdkScreen *screen;
-
+  GdkDevice *device;
   GdkWindow *window;
   GtkWidget *tooltip_widget;
   GtkTooltip *tooltip;
@@ -1095,10 +984,10 @@ gtk_tooltip_show_tooltip (GdkDisplay *display)
     {
       x = y = -1;
       tooltip_widget = tooltip->keyboard_widget;
+      device = NULL;
     }
   else
     {
-      GdkDevice *device;
       gint tx, ty;
 
       window = tooltip->last_window;
@@ -1111,8 +1000,6 @@ gtk_tooltip_show_tooltip (GdkDisplay *display)
       gdk_window_get_device_position (window, device, &x, &y, NULL);
 
       gdk_window_get_root_coords (window, x, y, &tx, &ty);
-      tooltip->last_x = tx;
-      tooltip->last_y = ty;
 
       tooltip_widget = _gtk_widget_find_at_coords (window, x, y, &x, &y);
     }
@@ -1147,7 +1034,7 @@ gtk_tooltip_show_tooltip (GdkDisplay *display)
                         G_CALLBACK (gtk_tooltip_display_closed), tooltip);
     }
 
-  gtk_tooltip_position (tooltip, display, tooltip_widget);
+  gtk_tooltip_position (tooltip, display, tooltip_widget, device);
 
   /* Now a tooltip is visible again on the display, make sure browse
    * mode is enabled.
@@ -1390,6 +1277,22 @@ _gtk_tooltip_hide (GtkWidget *widget)
     gtk_tooltip_hide_tooltip (tooltip);
 }
 
+void
+_gtk_tooltip_hide_in_display (GdkDisplay *display)
+{
+  GtkTooltip *tooltip;
+
+  if (!display)
+    return;
+
+  tooltip = g_object_get_qdata (G_OBJECT (display), quark_current_tooltip);
+
+  if (!tooltip || !GTK_TOOLTIP_VISIBLE (tooltip))
+    return;
+
+  gtk_tooltip_hide_tooltip (tooltip);
+}
+
 static gboolean
 tooltips_enabled (GdkEvent *event)
 {
@@ -1422,8 +1325,7 @@ static void
 gtk_tooltip_handle_event_internal (GdkEvent *event)
 {
   gint x, y;
-  gboolean return_value = FALSE;
-  GtkWidget *has_tooltip_widget = NULL;
+  GtkWidget *has_tooltip_widget;
   GdkDisplay *display;
   GtkTooltip *current_tooltip;
 
@@ -1439,6 +1341,8 @@ gtk_tooltip_handle_event_internal (GdkEvent *event)
 
   if (current_tooltip && current_tooltip->keyboard_mode_enabled)
     {
+      gboolean return_value;
+
       has_tooltip_widget = current_tooltip->keyboard_widget;
       if (!has_tooltip_widget)
 	return;
@@ -1491,12 +1395,9 @@ gtk_tooltip_handle_event_internal (GdkEvent *event)
 	    tip_area_set = current_tooltip->tip_area_set;
 	    tip_area = current_tooltip->tip_area;
 
-	    return_value = gtk_tooltip_run_requery (&has_tooltip_widget,
-						    current_tooltip,
-						    &x, &y);
-
-	    /* Requested to be hidden? */
-	    hide_tooltip = !return_value;
+	    gtk_tooltip_run_requery (&has_tooltip_widget,
+                                     current_tooltip,
+                                     &x, &y);
 
 	    /* Leave notify should override the query function */
 	    hide_tooltip = (event->type == GDK_LEAVE_NOTIFY);
