@@ -17,11 +17,16 @@
 
 #include "config.h"
 
+#include "gtkwidgetpath.h"
+
 #include <string.h>
 
-#include "gtkwidget.h"
-#include "gtkwidgetpath.h"
+#include "gtkcssnodedeclarationprivate.h"
+#include "gtkprivate.h"
 #include "gtkstylecontextprivate.h"
+#include "gtktypebuiltins.h"
+#include "gtkwidget.h"
+#include "gtkwidgetpathprivate.h"
 
 /**
  * SECTION:gtkwidgetpath
@@ -40,9 +45,10 @@
  * updated on widget hierarchy changes.
  *
  * The widget path generation is generally simple:
- * <example>
- * <title>Defining a button within a window</title>
- * <programlisting>
+ *
+ * ## Defining a button within a window
+ *
+ * |[<!-- language="C" -->
  * {
  *   GtkWidgetPath *path;
  *
@@ -50,16 +56,15 @@
  *   gtk_widget_path_append_type (path, GTK_TYPE_WINDOW);
  *   gtk_widget_path_append_type (path, GTK_TYPE_BUTTON);
  * }
- * </programlisting>
- * </example>
+ * ]|
  *
  * Although more complex information, such as widget names, or
  * different classes (property that may be used by other widget
  * types) and intermediate regions may be included:
  *
- * <example>
- * <title>Defining the first tab widget in a notebook</title>
- * <programlisting>
+ * ## Defining the first tab widget in a notebook
+ *
+ * |[<!-- language="C" -->
  * {
  *   GtkWidgetPath *path;
  *   guint pos;
@@ -72,8 +77,7 @@
  *   pos = gtk_widget_path_append_type (path, GTK_TYPE_LABEL);
  *   gtk_widget_path_iter_set_name (path, pos, "first tab label");
  * }
- * </programlisting>
- * </example>
+ * ]|
  *
  * All this information will be used to match the style information
  * that applies to the described widget.
@@ -87,17 +91,14 @@ typedef struct GtkPathElement GtkPathElement;
 
 struct GtkPathElement
 {
-  GType type;
-  GQuark name;
+  GtkCssNodeDeclaration *decl;
   guint sibling_index;
-  GHashTable *regions;
-  GArray *classes;
   GtkWidgetPath *siblings;
 };
 
 struct _GtkWidgetPath
 {
-  volatile guint ref_count;
+  guint ref_count;
 
   GArray *elems; /* First element contains the described widget */
 };
@@ -129,29 +130,10 @@ gtk_path_element_copy (GtkPathElement       *dest,
 {
   memset (dest, 0, sizeof (GtkPathElement));
 
-  dest->type = src->type;
-  dest->name = src->name;
+  dest->decl = gtk_css_node_declaration_ref (src->decl);
   if (src->siblings)
     dest->siblings = gtk_widget_path_ref (src->siblings);
   dest->sibling_index = src->sibling_index;
-
-  if (src->regions)
-    {
-      GHashTableIter iter;
-      gpointer key, value;
-
-      g_hash_table_iter_init (&iter, src->regions);
-      dest->regions = g_hash_table_new (NULL, NULL);
-
-      while (g_hash_table_iter_next (&iter, &key, &value))
-        g_hash_table_insert (dest->regions, key, value);
-    }
-
-  if (src->classes)
-    {
-      dest->classes = g_array_new (FALSE, FALSE, sizeof (GQuark));
-      g_array_append_vals (dest->classes, src->classes->data, src->classes->len);
-    }
 }
 
 /**
@@ -170,7 +152,7 @@ gtk_widget_path_copy (const GtkWidgetPath *path)
   GtkWidgetPath *new_path;
   guint i;
 
-  g_return_val_if_fail (path != NULL, NULL);
+  gtk_internal_return_val_if_fail (path != NULL, NULL);
 
   new_path = gtk_widget_path_new ();
 
@@ -202,9 +184,9 @@ gtk_widget_path_copy (const GtkWidgetPath *path)
 GtkWidgetPath *
 gtk_widget_path_ref (GtkWidgetPath *path)
 {
-  g_return_val_if_fail (path != NULL, path);
+  gtk_internal_return_val_if_fail (path != NULL, path);
 
-  g_atomic_int_add (&path->ref_count, 1);
+  path->ref_count += 1;
 
   return path;
 }
@@ -223,9 +205,10 @@ gtk_widget_path_unref (GtkWidgetPath *path)
 {
   guint i;
 
-  g_return_if_fail (path != NULL);
+  gtk_internal_return_if_fail (path != NULL);
 
-  if (!g_atomic_int_dec_and_test (&path->ref_count))
+  path->ref_count -= 1;
+  if (path->ref_count > 0)
     return;
 
   for (i = 0; i < path->elems->len; i++)
@@ -234,12 +217,7 @@ gtk_widget_path_unref (GtkWidgetPath *path)
 
       elem = &g_array_index (path->elems, GtkPathElement, i);
 
-      if (elem->regions)
-        g_hash_table_destroy (elem->regions);
-
-      if (elem->classes)
-        g_array_free (elem->classes, TRUE);
-
+      gtk_css_node_declaration_unref (elem->decl);
       if (elem->siblings)
         gtk_widget_path_unref (elem->siblings);
     }
@@ -260,7 +238,7 @@ gtk_widget_path_unref (GtkWidgetPath *path)
 void
 gtk_widget_path_free (GtkWidgetPath *path)
 {
-  g_return_if_fail (path != NULL);
+  gtk_internal_return_if_fail (path != NULL);
 
   gtk_widget_path_unref (path);
 }
@@ -279,7 +257,7 @@ gtk_widget_path_free (GtkWidgetPath *path)
 gint
 gtk_widget_path_length (const GtkWidgetPath *path)
 {
-  g_return_val_if_fail (path != NULL, 0);
+  gtk_internal_return_val_if_fail (path != NULL, 0);
 
   return path->elems->len;
 }
@@ -303,75 +281,92 @@ char *
 gtk_widget_path_to_string (const GtkWidgetPath *path)
 {
   GString *string;
-  guint i, j;
+  guint i, j, n;
 
-  g_return_val_if_fail (path != NULL, NULL);
+  gtk_internal_return_val_if_fail (path != NULL, NULL);
 
   string = g_string_new ("");
 
   for (i = 0; i < path->elems->len; i++)
     {
       GtkPathElement *elem;
+      GtkStateFlags state;
+      const GQuark *classes;
+      GList *list, *regions;
 
       elem = &g_array_index (path->elems, GtkPathElement, i);
 
       if (i > 0)
         g_string_append_c (string, ' ');
 
-      g_string_append (string, g_type_name (elem->type));
+      if (gtk_css_node_declaration_get_name (elem->decl))
+        g_string_append (string, gtk_css_node_declaration_get_name (elem->decl));
+      else
+        g_string_append (string, g_type_name (gtk_css_node_declaration_get_type (elem->decl)));
 
-      if (elem->name)
+      if (gtk_css_node_declaration_get_id (elem->decl))
         {
           g_string_append_c (string, '(');
-          g_string_append (string, g_quark_to_string (elem->name));
+          g_string_append (string, gtk_css_node_declaration_get_id (elem->decl));
           g_string_append_c (string, ')');
         }
 
+      state = gtk_css_node_declaration_get_state (elem->decl);
+      if (state)
+        {
+          GFlagsClass *fclass;
+
+          fclass = g_type_class_ref (GTK_TYPE_STATE_FLAGS);
+          for (j = 0; j < fclass->n_values; j++)
+            {
+              if (state & fclass->values[j].value)
+                {
+                  g_string_append_c (string, ':');
+                  g_string_append (string, fclass->values[j].value_nick);
+                }
+            }
+          g_type_class_unref (fclass);
+        }
 
       if (elem->siblings)
         g_string_append_printf (string, "[%d/%d]",
                                 elem->sibling_index + 1,
                                 gtk_widget_path_length (elem->siblings));
 
-      if (elem->classes)
+      classes = gtk_css_node_declaration_get_classes (elem->decl, &n);
+      for (j = 0; j < n; j++)
         {
-          for (j = 0; j < elem->classes->len; j++)
-            {
-              g_string_append_c (string, '.');
-              g_string_append (string, g_quark_to_string (g_array_index (elem->classes, GQuark, j)));
-            }
+          g_string_append_c (string, '.');
+          g_string_append (string, g_quark_to_string (classes[j]));
         }
 
-      if (elem->regions)
+      regions = gtk_css_node_declaration_list_regions (elem->decl);
+      for (list = regions; list; list = list->next)
         {
-          GHashTableIter iter;
-          gpointer key, value;
+          static const char *flag_names[] = {
+            "even",
+            "odd",
+            "first",
+            "last",
+            "only",
+            "sorted"
+          };
+          GtkRegionFlags flags;
+          GQuark region = GPOINTER_TO_UINT (regions->data);
 
-          g_hash_table_iter_init (&iter, elem->regions);
-          while (g_hash_table_iter_next (&iter, &key, &value))
+          gtk_css_node_declaration_has_region (elem->decl, region, &flags);
+          g_string_append_c (string, ' ');
+          g_string_append (string, g_quark_to_string (region));
+          for (j = 0; j < G_N_ELEMENTS(flag_names); j++)
             {
-              GtkRegionFlags flags = GPOINTER_TO_UINT (value);
-              static const char *flag_names[] = {
-                "even",
-                "odd",
-                "first",
-                "last",
-                "only",
-                "sorted"
-              };
-
-              g_string_append_c (string, ' ');
-              g_string_append (string, g_quark_to_string (GPOINTER_TO_UINT (key)));
-              for (j = 0; j < G_N_ELEMENTS(flag_names); j++)
+              if (flags & (1 << j))
                 {
-                  if (flags & (1 << j))
-                    {
-                      g_string_append_c (string, ':');
-                      g_string_append (string, flag_names[j]);
-                    }
+                  g_string_append_c (string, ':');
+                  g_string_append (string, flag_names[j]);
                 }
             }
         }
+      g_list_free (regions);
     }
 
   return g_string_free (string, FALSE);
@@ -390,11 +385,13 @@ void
 gtk_widget_path_prepend_type (GtkWidgetPath *path,
                               GType          type)
 {
-  GtkPathElement new = { 0 };
+  GtkPathElement new = { NULL };
 
-  g_return_if_fail (path != NULL);
+  gtk_internal_return_if_fail (path != NULL);
+
+  new.decl = gtk_css_node_declaration_new ();
+  gtk_css_node_declaration_set_type (&new.decl, type);
 
-  new.type = type;
   g_array_prepend_val (path->elems, new);
 }
 
@@ -413,11 +410,12 @@ gint
 gtk_widget_path_append_type (GtkWidgetPath *path,
                              GType          type)
 {
-  GtkPathElement new = { 0 };
+  GtkPathElement new = { NULL };
 
-  g_return_val_if_fail (path != NULL, 0);
+  gtk_internal_return_val_if_fail (path != NULL, 0);
 
-  new.type = type;
+  new.decl = gtk_css_node_declaration_new ();
+  gtk_css_node_declaration_set_type (&new.decl, type);
   g_array_append_val (path->elems, new);
 
   return path->elems->len - 1;
@@ -451,9 +449,9 @@ gtk_widget_path_append_with_siblings (GtkWidgetPath *path,
 {
   GtkPathElement new;
 
-  g_return_val_if_fail (path != NULL, 0);
-  g_return_val_if_fail (siblings != NULL, 0);
-  g_return_val_if_fail (sibling_index < gtk_widget_path_length (siblings), 0);
+  gtk_internal_return_val_if_fail (path != NULL, 0);
+  gtk_internal_return_val_if_fail (siblings != NULL, 0);
+  gtk_internal_return_val_if_fail (sibling_index < gtk_widget_path_length (siblings), 0);
 
   gtk_path_element_copy (&new, &g_array_index (siblings->elems, GtkPathElement, sibling_index));
   new.siblings = gtk_widget_path_ref (siblings);
@@ -479,8 +477,8 @@ gtk_widget_path_iter_get_siblings (const GtkWidgetPath *path,
 {
   GtkPathElement *elem;
 
-  g_return_val_if_fail (path != NULL, G_TYPE_INVALID);
-  g_return_val_if_fail (path->elems->len != 0, G_TYPE_INVALID);
+  gtk_internal_return_val_if_fail (path != NULL, NULL);
+  gtk_internal_return_val_if_fail (path->elems->len != 0, NULL);
 
   if (pos < 0 || pos >= path->elems->len)
     pos = path->elems->len - 1;
@@ -507,8 +505,8 @@ gtk_widget_path_iter_get_sibling_index (const GtkWidgetPath *path,
 {
   GtkPathElement *elem;
 
-  g_return_val_if_fail (path != NULL, G_TYPE_INVALID);
-  g_return_val_if_fail (path->elems->len != 0, G_TYPE_INVALID);
+  gtk_internal_return_val_if_fail (path != NULL, G_TYPE_INVALID);
+  gtk_internal_return_val_if_fail (path->elems->len != 0, G_TYPE_INVALID);
 
   if (pos < 0 || pos >= path->elems->len)
     pos = path->elems->len - 1;
@@ -517,6 +515,65 @@ gtk_widget_path_iter_get_sibling_index (const GtkWidgetPath *path,
   return elem->sibling_index;
 }
 
+/**
+ * gtk_widget_path_iter_get_object_name:
+ * @path: a #GtkWidgetPath
+ * @pos: position to get the object name for, -1 for the path head
+ *
+ * Returns the object name that is at position @pos in the widget
+ * hierarchy defined in @path.
+ *
+ * Returns: (nullable): the name or %NULL
+ *
+ * Since: 3.20
+ **/
+const char *
+gtk_widget_path_iter_get_object_name (const GtkWidgetPath *path,
+                                      gint                 pos)
+{
+  GtkPathElement *elem;
+
+  gtk_internal_return_val_if_fail (path != NULL, NULL);
+  gtk_internal_return_val_if_fail (path->elems->len != 0, NULL);
+
+  if (pos < 0 || pos >= path->elems->len)
+    pos = path->elems->len - 1;
+
+  elem = &g_array_index (path->elems, GtkPathElement, pos);
+  return gtk_css_node_declaration_get_name (elem->decl);
+}
+
+/**
+ * gtk_widget_path_iter_set_object_name:
+ * @path: a #GtkWidgetPath
+ * @pos: position to modify, -1 for the path head
+ * @name: (allow-none): object name to set or %NULL to unset
+ *
+ * Sets the object name for a given position in the widget hierarchy
+ * defined by @path.
+ *
+ * When set, the object name overrides the object type when matching
+ * CSS.
+ *
+ * Since: 3.20
+ **/
+void
+gtk_widget_path_iter_set_object_name (GtkWidgetPath *path,
+                                      gint           pos,
+                                      const char    *name)
+{
+  GtkPathElement *elem;
+
+  gtk_internal_return_if_fail (path != NULL);
+  gtk_internal_return_if_fail (path->elems->len != 0);
+
+  if (pos < 0 || pos >= path->elems->len)
+    pos = path->elems->len - 1;
+
+  elem = &g_array_index (path->elems, GtkPathElement, pos);
+  gtk_css_node_declaration_set_name (&elem->decl, g_intern_string (name));
+}
+
 /**
  * gtk_widget_path_iter_get_object_type:
  * @path: a #GtkWidgetPath
@@ -535,14 +592,14 @@ gtk_widget_path_iter_get_object_type (const GtkWidgetPath *path,
 {
   GtkPathElement *elem;
 
-  g_return_val_if_fail (path != NULL, G_TYPE_INVALID);
-  g_return_val_if_fail (path->elems->len != 0, G_TYPE_INVALID);
+  gtk_internal_return_val_if_fail (path != NULL, G_TYPE_INVALID);
+  gtk_internal_return_val_if_fail (path->elems->len != 0, G_TYPE_INVALID);
 
   if (pos < 0 || pos >= path->elems->len)
     pos = path->elems->len - 1;
 
   elem = &g_array_index (path->elems, GtkPathElement, pos);
-  return elem->type;
+  return gtk_css_node_declaration_get_type (elem->decl);
 }
 
 /**
@@ -563,14 +620,88 @@ gtk_widget_path_iter_set_object_type (GtkWidgetPath *path,
 {
   GtkPathElement *elem;
 
-  g_return_if_fail (path != NULL);
-  g_return_if_fail (path->elems->len != 0);
+  gtk_internal_return_if_fail (path != NULL);
+  gtk_internal_return_if_fail (path->elems->len != 0);
 
   if (pos < 0 || pos >= path->elems->len)
     pos = path->elems->len - 1;
 
   elem = &g_array_index (path->elems, GtkPathElement, pos);
-  elem->type = type;
+  gtk_css_node_declaration_set_type (&elem->decl, type);
+}
+
+/**
+ * gtk_widget_path_iter_get_state:
+ * @path: a #GtkWidgetPath
+ * @pos: position to get the state for, -1 for the path head
+ *
+ * Returns the state flags corresponding to the widget found at
+ * the position @pos in the widget hierarchy defined by
+ * @path
+ *
+ * Returns: The state flags
+ *
+ * Since: 3.14
+ **/
+GtkStateFlags
+gtk_widget_path_iter_get_state (const GtkWidgetPath *path,
+                                gint                 pos)
+{
+  GtkPathElement *elem;
+
+  gtk_internal_return_val_if_fail (path != NULL, 0);
+  gtk_internal_return_val_if_fail (path->elems->len != 0, 0);
+
+  if (pos < 0 || pos >= path->elems->len)
+    pos = path->elems->len - 1;
+
+  elem = &g_array_index (path->elems, GtkPathElement, pos);
+  return gtk_css_node_declaration_get_state (elem->decl);
+}
+
+/**
+ * gtk_widget_path_iter_set_state:
+ * @path: a #GtkWidgetPath
+ * @pos: position to modify, -1 for the path head
+ * @state: state flags
+ *
+ * Sets the widget name for the widget found at position @pos
+ * in the widget hierarchy defined by @path.
+ *
+ * If you want to update just a single state flag, you need to do
+ * this manually, as this function updates all state flags.
+ *
+ * ## Setting a flag
+ *
+ * |[<!-- language="C" -->
+ * gtk_widget_path_iter_set_state (path, pos, gtk_widget_path_iter_get_state (path, pos) | flag);
+ * ]|
+ *
+ * ## Unsetting a flag
+ *
+ * |[<!-- language="C" -->
+ * gtk_widget_path_iter_set_state (path, pos, gtk_widget_path_iter_get_state (path, pos) & ~flag);
+ * ]|
+ *
+ *
+ * Since: 3.14
+ **/
+void
+gtk_widget_path_iter_set_state (GtkWidgetPath *path,
+                                gint           pos,
+                                GtkStateFlags  state)
+{
+  GtkPathElement *elem;
+
+  gtk_internal_return_if_fail (path != NULL);
+  gtk_internal_return_if_fail (path->elems->len != 0);
+
+  if (pos < 0 || pos >= path->elems->len)
+    pos = path->elems->len - 1;
+
+  elem = &g_array_index (path->elems, GtkPathElement, pos);
+
+  gtk_css_node_declaration_set_state (&elem->decl, state);
 }
 
 /**
@@ -582,7 +713,7 @@ gtk_widget_path_iter_set_object_type (GtkWidgetPath *path,
  * the position @pos in the widget hierarchy defined by
  * @path
  *
- * Returns: The widget name, or %NULL if none was set.
+ * Returns: (nullable): The widget name, or %NULL if none was set.
  **/
 const gchar *
 gtk_widget_path_iter_get_name (const GtkWidgetPath *path,
@@ -590,14 +721,14 @@ gtk_widget_path_iter_get_name (const GtkWidgetPath *path,
 {
   GtkPathElement *elem;
 
-  g_return_val_if_fail (path != NULL, NULL);
-  g_return_val_if_fail (path->elems->len != 0, NULL);
+  gtk_internal_return_val_if_fail (path != NULL, NULL);
+  gtk_internal_return_val_if_fail (path->elems->len != 0, NULL);
 
   if (pos < 0 || pos >= path->elems->len)
     pos = path->elems->len - 1;
 
   elem = &g_array_index (path->elems, GtkPathElement, pos);
-  return g_quark_to_string (elem->name);
+  return gtk_css_node_declaration_get_id (elem->decl);
 }
 
 /**
@@ -618,16 +749,16 @@ gtk_widget_path_iter_set_name (GtkWidgetPath *path,
 {
   GtkPathElement *elem;
 
-  g_return_if_fail (path != NULL);
-  g_return_if_fail (path->elems->len != 0);
-  g_return_if_fail (name != NULL);
+  gtk_internal_return_if_fail (path != NULL);
+  gtk_internal_return_if_fail (path->elems->len != 0);
+  gtk_internal_return_if_fail (name != NULL);
 
   if (pos < 0 || pos >= path->elems->len)
     pos = path->elems->len - 1;
 
   elem = &g_array_index (path->elems, GtkPathElement, pos);
 
-  elem->name = g_quark_from_string (name);
+  gtk_css_node_declaration_set_id (&elem->decl, g_intern_string (name));
 }
 
 /**
@@ -637,7 +768,7 @@ gtk_widget_path_iter_set_name (GtkWidgetPath *path,
  * @qname: widget name as a #GQuark
  *
  * See gtk_widget_path_iter_has_name(). This is a version
- * that operates on #GQuark<!-- -->s.
+ * that operates on #GQuarks.
  *
  * Returns: %TRUE if the widget at @pos has this name
  *
@@ -648,18 +779,11 @@ gtk_widget_path_iter_has_qname (const GtkWidgetPath *path,
                                 gint                 pos,
                                 GQuark               qname)
 {
-  GtkPathElement *elem;
+  gtk_internal_return_val_if_fail (path != NULL, FALSE);
+  gtk_internal_return_val_if_fail (path->elems->len != 0, FALSE);
+  gtk_internal_return_val_if_fail (qname != 0, FALSE);
 
-  g_return_val_if_fail (path != NULL, FALSE);
-  g_return_val_if_fail (path->elems->len != 0, FALSE);
-  g_return_val_if_fail (qname != 0, FALSE);
-
-  if (pos < 0 || pos >= path->elems->len)
-    pos = path->elems->len - 1;
-
-  elem = &g_array_index (path->elems, GtkPathElement, pos);
-
-  return (elem->name == qname);
+  return gtk_widget_path_iter_has_name (path, pos, g_quark_to_string (qname));
 }
 
 /**
@@ -680,20 +804,18 @@ gtk_widget_path_iter_has_name (const GtkWidgetPath *path,
                                gint                 pos,
                                const gchar         *name)
 {
-  GQuark qname;
+  GtkPathElement *elem;
 
-  g_return_val_if_fail (path != NULL, FALSE);
-  g_return_val_if_fail (path->elems->len != 0, FALSE);
+  gtk_internal_return_val_if_fail (path != NULL, FALSE);
+  gtk_internal_return_val_if_fail (path->elems->len != 0, FALSE);
 
   if (pos < 0 || pos >= path->elems->len)
     pos = path->elems->len - 1;
 
-  qname = g_quark_try_string (name);
+  name = g_intern_string (name);
+  elem = &g_array_index (path->elems, GtkPathElement, pos);
 
-  if (qname == 0)
-    return FALSE;
-
-  return gtk_widget_path_iter_has_qname (path, pos, qname);
+  return gtk_css_node_declaration_get_id (elem->decl) == name;
 }
 
 /**
@@ -713,46 +835,26 @@ gtk_widget_path_iter_add_class (GtkWidgetPath *path,
                                 gint           pos,
                                 const gchar   *name)
 {
-  GtkPathElement *elem;
-  gboolean added = FALSE;
-  GQuark qname;
-  guint i;
-
-  g_return_if_fail (path != NULL);
-  g_return_if_fail (path->elems->len != 0);
-  g_return_if_fail (name != NULL);
+  gtk_internal_return_if_fail (path != NULL);
+  gtk_internal_return_if_fail (path->elems->len != 0);
+  gtk_internal_return_if_fail (name != NULL);
 
   if (pos < 0 || pos >= path->elems->len)
     pos = path->elems->len - 1;
 
+  gtk_widget_path_iter_add_qclass (path, pos, g_quark_from_string (name));
+}
+
+void
+gtk_widget_path_iter_add_qclass (GtkWidgetPath *path,
+                                 gint           pos,
+                                 GQuark         qname)
+{
+  GtkPathElement *elem;
+
   elem = &g_array_index (path->elems, GtkPathElement, pos);
-  qname = g_quark_from_string (name);
 
-  if (!elem->classes)
-    elem->classes = g_array_new (FALSE, FALSE, sizeof (GQuark));
-
-  for (i = 0; i < elem->classes->len; i++)
-    {
-      GQuark quark;
-
-      quark = g_array_index (elem->classes, GQuark, i);
-
-      if (qname == quark)
-        {
-          /* Already there */
-          added = TRUE;
-          break;
-        }
-      if (qname < quark)
-        {
-          g_array_insert_val (elem->classes, i, qname);
-          added = TRUE;
-          break;
-        }
-    }
-
-  if (!added)
-    g_array_append_val (elem->classes, qname);
+  gtk_css_node_declaration_add_class (&elem->decl, qname);
 }
 
 /**
@@ -773,39 +875,20 @@ gtk_widget_path_iter_remove_class (GtkWidgetPath *path,
 {
   GtkPathElement *elem;
   GQuark qname;
-  guint i;
 
-  g_return_if_fail (path != NULL);
-  g_return_if_fail (path->elems->len != 0);
-  g_return_if_fail (name != NULL);
+  gtk_internal_return_if_fail (path != NULL);
+  gtk_internal_return_if_fail (path->elems->len != 0);
+  gtk_internal_return_if_fail (name != NULL);
 
   if (pos < 0 || pos >= path->elems->len)
     pos = path->elems->len - 1;
 
+  elem = &g_array_index (path->elems, GtkPathElement, pos);
   qname = g_quark_try_string (name);
-
   if (qname == 0)
     return;
 
-  elem = &g_array_index (path->elems, GtkPathElement, pos);
-
-  if (!elem->classes)
-    return;
-
-  for (i = 0; i < elem->classes->len; i++)
-    {
-      GQuark quark;
-
-      quark = g_array_index (elem->classes, GQuark, i);
-
-      if (quark > qname)
-        break;
-      else if (quark == qname)
-        {
-          g_array_remove_index (elem->classes, i);
-          break;
-        }
-    }
+  gtk_css_node_declaration_remove_class (&elem->decl, qname);
 }
 
 /**
@@ -824,19 +907,15 @@ gtk_widget_path_iter_clear_classes (GtkWidgetPath *path,
 {
   GtkPathElement *elem;
 
-  g_return_if_fail (path != NULL);
-  g_return_if_fail (path->elems->len != 0);
+  gtk_internal_return_if_fail (path != NULL);
+  gtk_internal_return_if_fail (path->elems->len != 0);
 
   if (pos < 0 || pos >= path->elems->len)
     pos = path->elems->len - 1;
 
   elem = &g_array_index (path->elems, GtkPathElement, pos);
 
-  if (!elem->classes)
-    return;
-
-  if (elem->classes->len > 0)
-    g_array_remove_range (elem->classes, 0, elem->classes->len);
+  gtk_css_node_declaration_clear_classes (&elem->decl);
 }
 
 /**
@@ -860,25 +939,21 @@ gtk_widget_path_iter_list_classes (const GtkWidgetPath *path,
 {
   GtkPathElement *elem;
   GSList *list = NULL;
-  guint i;
+  const GQuark *classes;
+  guint i, n;
 
-  g_return_val_if_fail (path != NULL, NULL);
-  g_return_val_if_fail (path->elems->len != 0, NULL);
+  gtk_internal_return_val_if_fail (path != NULL, NULL);
+  gtk_internal_return_val_if_fail (path->elems->len != 0, NULL);
 
   if (pos < 0 || pos >= path->elems->len)
     pos = path->elems->len - 1;
 
   elem = &g_array_index (path->elems, GtkPathElement, pos);
+  classes = gtk_css_node_declaration_get_classes (elem->decl, &n);
 
-  if (!elem->classes)
-    return NULL;
-
-  for (i = 0; i < elem->classes->len; i++)
+  for (i = 0; i < n; i++)
     {
-      GQuark quark;
-
-      quark = g_array_index (elem->classes, GQuark, i);
-      list = g_slist_prepend (list, (gchar *) g_quark_to_string (quark));
+      list = g_slist_prepend (list, (gchar *) g_quark_to_string (classes[i]));
     }
 
   return g_slist_reverse (list);
@@ -891,7 +966,7 @@ gtk_widget_path_iter_list_classes (const GtkWidgetPath *path,
  * @qname: class name as a #GQuark
  *
  * See gtk_widget_path_iter_has_class(). This is a version that operates
- * with GQuark<!-- -->s.
+ * with GQuarks.
  *
  * Returns: %TRUE if the widget at @pos has the class defined.
  *
@@ -903,33 +978,17 @@ gtk_widget_path_iter_has_qclass (const GtkWidgetPath *path,
                                  GQuark               qname)
 {
   GtkPathElement *elem;
-  guint i;
 
-  g_return_val_if_fail (path != NULL, FALSE);
-  g_return_val_if_fail (path->elems->len != 0, FALSE);
-  g_return_val_if_fail (qname != 0, FALSE);
+  gtk_internal_return_val_if_fail (path != NULL, FALSE);
+  gtk_internal_return_val_if_fail (path->elems->len != 0, FALSE);
+  gtk_internal_return_val_if_fail (qname != 0, FALSE);
 
   if (pos < 0 || pos >= path->elems->len)
     pos = path->elems->len - 1;
 
   elem = &g_array_index (path->elems, GtkPathElement, pos);
 
-  if (!elem->classes)
-    return FALSE;
-
-  for (i = 0; i < elem->classes->len; i++)
-    {
-      GQuark quark;
-
-      quark = g_array_index (elem->classes, GQuark, i);
-
-      if (quark == qname)
-        return TRUE;
-      else if (quark > qname)
-        break;
-    }
-
-  return FALSE;
+  return gtk_css_node_declaration_has_class (elem->decl, qname);
 }
 
 /**
@@ -952,9 +1011,9 @@ gtk_widget_path_iter_has_class (const GtkWidgetPath *path,
 {
   GQuark qname;
 
-  g_return_val_if_fail (path != NULL, FALSE);
-  g_return_val_if_fail (path->elems->len != 0, FALSE);
-  g_return_val_if_fail (name != NULL, FALSE);
+  gtk_internal_return_val_if_fail (path != NULL, FALSE);
+  gtk_internal_return_val_if_fail (path->elems->len != 0, FALSE);
+  gtk_internal_return_val_if_fail (name != NULL, FALSE);
 
   if (pos < 0 || pos >= path->elems->len)
     pos = path->elems->len - 1;
@@ -978,10 +1037,12 @@ gtk_widget_path_iter_has_class (const GtkWidgetPath *path,
  * the hierarchy defined in @path. See
  * gtk_style_context_add_region().
  *
- * <note><para>Region names must only contain lowercase letters
- * and '-', starting always with a lowercase letter.</para></note>
+ * Region names must only contain lowercase letters
+ * and “-”, starting always with a lowercase letter.
  *
  * Since: 3.0
+ *
+ * Deprecated: 3.14: The use of regions is deprecated.
  **/
 void
 gtk_widget_path_iter_add_region (GtkWidgetPath  *path,
@@ -992,10 +1053,10 @@ gtk_widget_path_iter_add_region (GtkWidgetPath  *path,
   GtkPathElement *elem;
   GQuark qname;
 
-  g_return_if_fail (path != NULL);
-  g_return_if_fail (path->elems->len != 0);
-  g_return_if_fail (name != NULL);
-  g_return_if_fail (_gtk_style_context_check_region_name (name));
+  gtk_internal_return_if_fail (path != NULL);
+  gtk_internal_return_if_fail (path->elems->len != 0);
+  gtk_internal_return_if_fail (name != NULL);
+  gtk_internal_return_if_fail (_gtk_style_context_check_region_name (name));
 
   if (pos < 0 || pos >= path->elems->len)
     pos = path->elems->len - 1;
@@ -1003,12 +1064,7 @@ gtk_widget_path_iter_add_region (GtkWidgetPath  *path,
   elem = &g_array_index (path->elems, GtkPathElement, pos);
   qname = g_quark_from_string (name);
 
-  if (!elem->regions)
-    elem->regions = g_hash_table_new (NULL, NULL);
-
-  g_hash_table_insert (elem->regions,
-                       GUINT_TO_POINTER (qname),
-                       GUINT_TO_POINTER (flags));
+  gtk_css_node_declaration_add_region (&elem->decl, qname, flags);
 }
 
 /**
@@ -1021,6 +1077,8 @@ gtk_widget_path_iter_add_region (GtkWidgetPath  *path,
  * the hierarchy defined in @path.
  *
  * Since: 3.0
+ *
+ * Deprecated: 3.14: The use of regions is deprecated.
  **/
 void
 gtk_widget_path_iter_remove_region (GtkWidgetPath *path,
@@ -1030,22 +1088,19 @@ gtk_widget_path_iter_remove_region (GtkWidgetPath *path,
   GtkPathElement *elem;
   GQuark qname;
 
-  g_return_if_fail (path != NULL);
-  g_return_if_fail (path->elems->len != 0);
-  g_return_if_fail (name != NULL);
+  gtk_internal_return_if_fail (path != NULL);
+  gtk_internal_return_if_fail (path->elems->len != 0);
+  gtk_internal_return_if_fail (name != NULL);
 
   if (pos < 0 || pos >= path->elems->len)
     pos = path->elems->len - 1;
 
+  elem = &g_array_index (path->elems, GtkPathElement, pos);
   qname = g_quark_try_string (name);
-
   if (qname == 0)
     return;
 
-  elem = &g_array_index (path->elems, GtkPathElement, pos);
-
-  if (elem->regions)
-    g_hash_table_remove (elem->regions, GUINT_TO_POINTER (qname));
+  gtk_css_node_declaration_remove_region (&elem->decl, qname);
 }
 
 /**
@@ -1057,6 +1112,8 @@ gtk_widget_path_iter_remove_region (GtkWidgetPath *path,
  * hierarchy defined in @path.
  *
  * Since: 3.0
+ *
+ * Deprecated: 3.14: The use of regions is deprecated.
  **/
 void
 gtk_widget_path_iter_clear_regions (GtkWidgetPath *path,
@@ -1064,16 +1121,15 @@ gtk_widget_path_iter_clear_regions (GtkWidgetPath *path,
 {
   GtkPathElement *elem;
 
-  g_return_if_fail (path != NULL);
-  g_return_if_fail (path->elems->len != 0);
+  gtk_internal_return_if_fail (path != NULL);
+  gtk_internal_return_if_fail (path->elems->len != 0);
 
   if (pos < 0 || pos >= path->elems->len)
     pos = path->elems->len - 1;
 
   elem = &g_array_index (path->elems, GtkPathElement, pos);
 
-  if (elem->regions)
-    g_hash_table_remove_all (elem->regions);
+  gtk_css_node_declaration_clear_regions (&elem->decl);
 }
 
 /**
@@ -1090,36 +1146,31 @@ gtk_widget_path_iter_clear_regions (GtkWidgetPath *path,
  *          free the list itself.
  *
  * Since: 3.0
+ *
+ * Deprecated: 3.14: The use of regions is deprecated.
  **/
 GSList *
 gtk_widget_path_iter_list_regions (const GtkWidgetPath *path,
                                    gint                 pos)
 {
   GtkPathElement *elem;
-  GHashTableIter iter;
   GSList *list = NULL;
-  gpointer key;
+  GList *l, *wrong_list;
 
-  g_return_val_if_fail (path != NULL, NULL);
-  g_return_val_if_fail (path->elems->len != 0, NULL);
+  gtk_internal_return_val_if_fail (path != NULL, NULL);
+  gtk_internal_return_val_if_fail (path->elems->len != 0, NULL);
 
   if (pos < 0 || pos >= path->elems->len)
     pos = path->elems->len - 1;
 
   elem = &g_array_index (path->elems, GtkPathElement, pos);
 
-  if (!elem->regions)
-    return NULL;
-
-  g_hash_table_iter_init (&iter, elem->regions);
-
-  while (g_hash_table_iter_next (&iter, &key, NULL))
+  wrong_list = gtk_css_node_declaration_list_regions (elem->decl);
+  for (l = wrong_list; l; l = l->next)
     {
-      GQuark qname;
-
-      qname = GPOINTER_TO_UINT (key);
-      list = g_slist_prepend (list, (gchar *) g_quark_to_string (qname));
+      list = g_slist_prepend (list, (char *) g_quark_to_string (GPOINTER_TO_UINT (l->data)));
     }
+  g_list_free (wrong_list);
 
   return list;
 }
@@ -1132,11 +1183,13 @@ gtk_widget_path_iter_list_regions (const GtkWidgetPath *path,
  * @flags: (out): return location for the region flags
  *
  * See gtk_widget_path_iter_has_region(). This is a version that operates
- * with GQuark<!-- -->s.
+ * with GQuarks.
  *
  * Returns: %TRUE if the widget at @pos has the region defined.
  *
  * Since: 3.0
+ *
+ * Deprecated: 3.14: The use of regions is deprecated.
  **/
 gboolean
 gtk_widget_path_iter_has_qregion (const GtkWidgetPath *path,
@@ -1145,29 +1198,17 @@ gtk_widget_path_iter_has_qregion (const GtkWidgetPath *path,
                                   GtkRegionFlags      *flags)
 {
   GtkPathElement *elem;
-  gpointer value;
 
-  g_return_val_if_fail (path != NULL, FALSE);
-  g_return_val_if_fail (path->elems->len != 0, FALSE);
-  g_return_val_if_fail (qname != 0, FALSE);
+  gtk_internal_return_val_if_fail (path != NULL, FALSE);
+  gtk_internal_return_val_if_fail (path->elems->len != 0, FALSE);
+  gtk_internal_return_val_if_fail (qname != 0, FALSE);
 
   if (pos < 0 || pos >= path->elems->len)
     pos = path->elems->len - 1;
 
   elem = &g_array_index (path->elems, GtkPathElement, pos);
 
-  if (!elem->regions)
-    return FALSE;
-
-  if (!g_hash_table_lookup_extended (elem->regions,
-                                     GUINT_TO_POINTER (qname),
-                                     NULL, &value))
-    return FALSE;
-
-  if (flags)
-    *flags = GPOINTER_TO_UINT (value);
-
-  return TRUE;
+  return gtk_css_node_declaration_has_region (elem->decl, qname, flags);
 }
 
 /**
@@ -1183,6 +1224,8 @@ gtk_widget_path_iter_has_qregion (const GtkWidgetPath *path,
  * Returns: %TRUE if the class @name is defined for the widget at @pos
  *
  * Since: 3.0
+ *
+ * Deprecated: 3.14: The use of regions is deprecated.
  **/
 gboolean
 gtk_widget_path_iter_has_region (const GtkWidgetPath *path,
@@ -1192,9 +1235,9 @@ gtk_widget_path_iter_has_region (const GtkWidgetPath *path,
 {
   GQuark qname;
 
-  g_return_val_if_fail (path != NULL, FALSE);
-  g_return_val_if_fail (path->elems->len != 0, FALSE);
-  g_return_val_if_fail (name != NULL, FALSE);
+  gtk_internal_return_val_if_fail (path != NULL, FALSE);
+  gtk_internal_return_val_if_fail (path->elems->len != 0, FALSE);
+  gtk_internal_return_val_if_fail (name != NULL, FALSE);
 
   if (pos < 0 || pos >= path->elems->len)
     pos = path->elems->len - 1;
@@ -1204,7 +1247,9 @@ gtk_widget_path_iter_has_region (const GtkWidgetPath *path,
   if (qname == 0)
     return FALSE;
 
+G_GNUC_BEGIN_IGNORE_DEPRECATIONS
   return gtk_widget_path_iter_has_qregion (path, pos, qname, flags);
+G_GNUC_END_IGNORE_DEPRECATIONS
 }
 
 /**
@@ -1223,11 +1268,11 @@ gtk_widget_path_get_object_type (const GtkWidgetPath *path)
 {
   GtkPathElement *elem;
 
-  g_return_val_if_fail (path != NULL, G_TYPE_INVALID);
+  gtk_internal_return_val_if_fail (path != NULL, G_TYPE_INVALID);
 
   elem = &g_array_index (path->elems, GtkPathElement,
                          path->elems->len - 1);
-  return elem->type;
+  return gtk_css_node_declaration_get_type (elem->decl);
 }
 
 /**
@@ -1248,16 +1293,12 @@ gtk_widget_path_is_type (const GtkWidgetPath *path,
 {
   GtkPathElement *elem;
 
-  g_return_val_if_fail (path != NULL, FALSE);
+  gtk_internal_return_val_if_fail (path != NULL, FALSE);
 
   elem = &g_array_index (path->elems, GtkPathElement,
                          path->elems->len - 1);
 
-  if (elem->type == type ||
-      g_type_is_a (elem->type, type))
-    return TRUE;
-
-  return FALSE;
+  return g_type_is_a (gtk_css_node_declaration_get_type (elem->decl), type);
 }
 
 /**
@@ -1278,7 +1319,7 @@ gtk_widget_path_has_parent (const GtkWidgetPath *path,
 {
   guint i;
 
-  g_return_val_if_fail (path != NULL, FALSE);
+  gtk_internal_return_val_if_fail (path != NULL, FALSE);
 
   for (i = 0; i < path->elems->len - 1; i++)
     {
@@ -1286,8 +1327,7 @@ gtk_widget_path_has_parent (const GtkWidgetPath *path,
 
       elem = &g_array_index (path->elems, GtkPathElement, i);
 
-      if (elem->type == type ||
-          g_type_is_a (elem->type, type))
+      if (g_type_is_a (gtk_css_node_declaration_get_type (elem->decl), type))
         return TRUE;
     }
 
